do
	local itemTable = {}
	local slotLocations = {}
	local slotLevels = {}
	local usedItems = {}
	local excludeItems = {}
	-- List with a hash lookup, because I'm lazy
	local JEWELER_GEMS_LIST = { 42142, 36766, 42148, 42143, 42152, 42153, 42146, 42158, 42154, 42150, 42156, 42144, 42149, 36767, 42145, 42155, 42151, 42157 }
	local UNIQUE_GEMS_LIST = { 49110, 42702, 42701 }
	
	------------------------------------------------------------------------------------------------------------------------------
	------------------------------------------------------------------------------------------------------------------------------
	
	local JEWELER_GEMS, UNIQUE_GEMS = {}, {}
	for _, id in ipairs(JEWELER_GEMS_LIST) do JEWELER_GEMS[id] = true end	
	for _, id in ipairs(UNIQUE_GEMS_LIST) do UNIQUE_GEMS[id] = true end	
	
	local function numJCGems(location)
		local id, name, textureName, count, durability, maxDurability, invType, locked, start, duration, enable, setTooltip, gem1, gem2, gem3 = EquipmentManager_GetItemInfoByLocation(location)
		return (JEWELER_GEMS[gem1] and 1 or 0) + (JEWELER_GEMS[gem2] and 1 or 0) + (JEWELER_GEMS[gem3] and 1 or 0)
	end
	
	local function numUniqueGems(location)
		local id, name, textureName, count, durability, maxDurability, invType, locked, start, duration, enable, setTooltip, gem1, gem2, gem3 = EquipmentManager_GetItemInfoByLocation(location)	
		return (UNIQUE_GEMS[gem1] and 1 or 0) + (UNIQUE_GEMS[gem2] and 1 or 0) + (UNIQUE_GEMS[gem3] and 1 or 0)
	end	
	
	local function ilvlForLocation(location)
		local id = EquipmentManager_GetItemInfoByLocation(location)
		local _, _, _, itemLevel = GetItemInfo(id)
		return itemLevel - ((numJCGems(location) / 100.0) + (numUniqueGems(location) / 100.0 * 3))
	end
	-- Get the highest ilvl item for a slot. Prefers items without JC gems or Nightmare Tears
	local function getHighestILvlForSlot(slot, highestLevel, numJC, numUniq)
		numJC = numJC or 3
		numUniq = numUniq or 1
		if numJC < 0 then numJC = 0 end
		if numUniq < 0 then numUniq = 0 end
		
		wipe(itemTable)
		GetInventoryItemsForSlot(slot, itemTable)
		
		local highestLocation
		for location, id in pairs(itemTable) do
			
			local valid = numJCGems(location) <= numJC and numUniqueGems(location) <= numUniq
			local itemLevel = ilvlForLocation(location)
			
			if valid and itemLevel > highestLevel and not usedItems[location] and not excludeItems[location] then
				highestLevel = itemLevel
				highestLocation = location
			end
		end	
		return highestLevel, highestLocation
	end
	
	local function getInventoryItemLevels(i)
		slotLevels[i] = 0
		wipe(itemTable)
		GetInventoryItemsForSlot(i, itemTable)
	
		-- Establish the ilvl baseline with gear already being worn
		-- This makes worn items ineligible as replacements, and tallies the player's total ilevel
		for location, id in pairs(itemTable) do
			local player, bank, bags, slot = EquipmentManager_UnpackLocation(location)
			if player and not bank and not bags and slot == i then
				local itemLevel = ilvlForLocation(location)
				slotLevels[i] = itemLevel
				usedItems[location] = true
				return itemLevel
			end
		end
		
		return 0
	end	
	
	local function iterateGearChoices()
		wipe(slotLocations)
		wipe(slotLevels)
		wipe(usedItems)
	
		local initialLevel = 0
		for i = 1, 18 do
			if i ~= 4 and i ~= 17 then
				initialLevel = initialLevel + getInventoryItemLevels(i)
				-- Get the highest ilvl for this slot.
				slotLevels[i], slotLocations[i] = getHighestILvlForSlot(i, slotLevels[i])
			end
		end	
		return initialLevel
	end
	
	-- A rather stupid function, but it works. This doesn't actually solve the problem entirely - it just accepts the first 3 pieces of
	-- JC-gemmed gear. The reason is the solving the problem is rather complex; if you're over the gem limit, determining which pieces
	-- to drop can be non-trivial, as you have to assign an opportunity cost to each gem slotted, then determine which iteration of gear
	-- confers the lowest opportunity cost. This is a lazy implementation, but it'll keep the mod from breaking for people with JC gems.
	local deltas = {}
	local function checkInvalidJCGems()
		local totalJC, totalUnique = 0, 0
		for slot, location in pairs(slotLocations) do
			totalJC = totalJC + numJCGems(location)
			totalUnique = totalUnique + numUniqueGems(location)
			
			if totalJC > 3 or totalUnique > 1 then
				slotLocations[slot] = getHighestILvlForSlot(slot, 0, 3 - totalJC, 1 - totalUnique)
			end				
		end
		return valid
	end

	function EquipmentManager_EquipMaxILvl()
		wipe(excludeItems)
		
		local initialLevel = iterateGearChoices()
		checkInvalidJCGems()
		
		local totalLevels = 0
		for id, level in pairs(slotLevels) do
			totalLevels = totalLevels + level
		end

		local num = 0
		for slot, location in pairs(slotLocations) do
			local action = EquipmentManager_EquipItemByLocation(location, slot)
			if action then
				EquipmentManager_RunAction(action)
				num = num + 1
			end
		end
		ChatFrame1:AddMessage(("|cff00aaddMax gear iLvl:|r |cffffff00%s|r items equipped, old total iLvl: |cffffff00%s|r; new total iLvl: |cffffff00%s|r"):format(num, math.ceil(initialLevel), math.ceil(totalLevels)))
	end

	local function createMaxILvlButton(...)
		local f = CreateFrame("Button", "EquipmentManager_EquipMaxILvlButton", GearManagerDialog, "ItemButtonTemplate")
		f:SetPoint("TOPLEFT", GearManagerDialog, "TOPRIGHT", 0, -30)
		EquipmentManager_EquipMaxILvlButtonIconTexture:SetTexture("Interface\\Icons\\Spell_Shadow_TwistedFaith")
		f:SetScript("OnClick", EquipmentManager_EquipMaxILvl)
		local t = f:CreateFontString(nil, nil, "SystemFont_Tiny")
		t:SetPoint("BOTTOM")
		t:SetText("Max iLvl")
	end	

	createMaxILvlButton()
end